/*
 * last fixed: 2013.08.23.
 * by Wang Yi.
 * */
#ifndef __LEARNLAMDA_H_

#define __LEARNLAMDA_H_
#include <iostream>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <omp.h>
#include "Ncbi_nodes_dmp.h"
#include "GLOBAL.h"

using namespace std;
/*bool isSameFold(const string&p1,const string&p2){
	int idx1 = p1.length()-1, idx2 = p2.length()-1;
	while(idx1>0 && p1[idx1]!='/')--idx1;
	while(idx2>0 && p2[idx2]!='/')--idx2;
	if(idx1!=idx2)return false;
	for(;idx1>=0;--idx1)
		if(p1[idx1]!=p2[idx1])
			return false;
	return true;
}*/

bool pairCmp(const pair<double,bool>& t1,const pair<double,bool>& t2){
	if(t1.first==t2.first && t1.second!=t2.second)return !t1.second;
	return t1.first < t2.first;
}
double calLamda(const double tN_WEIGHT,const vector<double>& intra,const vector<double>& inter,int&OTP,int&OTN){
	vector<pair<double,bool> > V(inter.size()+intra.size());
	for(int i=0;i<inter.size();++i)
		V[i] = make_pair(inter[i],true);
	for(int i=0;i<intra.size();++i)
		V[i+inter.size()] = make_pair(intra[i],false);
	sort(V.begin(),V.end(),pairCmp);
	double max = tN_WEIGHT*intra.size();int maxid = -1;
	int TP = 0, TN = intra.size();
	OTP = TP; OTN = TN;
	for(int i=0;i<V.size();++i){
		if(V[i].second){
			++TP;
			if(TP+TN*tN_WEIGHT>max){
				max = TP+TN*tN_WEIGHT;
				maxid = i;
				OTP = TP;
				OTN = TN;
			}
		}
		else  --TN;
	}
//	if(maxid==-1)return -1;
//	else if(maxid==V.size()-1)return 1e99;
	if(V.size()==0)return 0;
	else if(maxid==-1)return V[0].first;
	else if(maxid==V.size()-1)return V[maxid].first;
	else return (V[maxid].first + V[maxid+1].first)/2;
}

void getLamda(vector<vector<int> >&Taxon, vector<vector<double> >&Mtx, map<int,vector<double> >&SidLamda, map<int,vector<double> > &SidPrec, map<int,vector<pair<double,double> > > &SidNormal,map<int,vector<int> >&SidNormalPrior){
	int N = Mtx.size();
 	const int L7[] = {4,8,12,17,21,24,28};
	const int NL = sizeof(L7)/sizeof(int);
	cerr << "NL :" << NL << endl;
	vector<int>TP(NL,0);
	vector<int>TN(NL,0);
	vector<int>PN(NL,0);
	for(int i=0;i<N;++i){
		vector<double> Intra[NL+1];
		for(int j=0;j<N;++j){
			int k = 0;
			for(;k<NL && Taxon[i][L7[k]]!=Taxon[j][L7[k]];++k);
		//TODO:test:
		//	if(k>3 || Mtx[i][j]<1)continue;
	//		if(k>3 || Mtx[i][j]<1)continue;
	//		if(k>3)continue;
			Intra[k].push_back(Mtx[i][j]/Mtx[j][j]);
		}
		///////////////////
		//calculate normal distribution
		vector<pair<double,double> >normalPara(NL+1);
		vector<int>normalPrior(NL+1);
		for(int j=0;j<=NL;++j){
			if(Intra[j].size()==0){
				normalPara[j] = make_pair(0,0);
				continue;
			}
			double sumX = 0, sumXX = 0;
			for(vector<double>::const_iterator itr=Intra[j].begin();itr!=Intra[j].end();++itr){
				sumX += *itr;
				sumXX += (*itr)*(*itr);
			}
			sumX /= Intra[j].size();
			normalPara[j] = make_pair(sumX,sqrt(sumXX/Intra[j].size()-sumX*sumX));
			normalPrior[j] = Intra[j].size();
		}
		////////////////////
		vector<double> lamdas(NL);
		vector<double> precs(NL);
		for(int j=0;j<NL;++j){
			//Test
	/*		double tN_WEIGHT = TN_WEIGHT*(NL-j);
			int x=tN_WEIGHT;
			if(x%2==0)--tN_WEIGHT;
			tN_WEIGHT = TN_WEIGHT;*/
			/*
			double tN_WEIGHT = 5-2*j;
			if(tN_WEIGHT < 0)tN_WEIGHT = 10;*/
			///////////////

			int OTP=0,OTN=0;
/*			vector<double>v1,v2;
			for(int k=0;k<=j;++k)
				for(vector<double>::const_iterator itr=Intra[k].begin();itr!=Intra[k].end();++itr)
					v1.push_back(*itr);
			for(int k=j+1;k<=NL;++k)
				for(vector<double>::const_iterator itr=Intra[k].begin();itr!=Intra[k].end();++itr)
					v2.push_back(*itr);*/
//			lamdas[j]=calLamda(TN_WEIGHT,Intra[j], Intra[j+1],OTP,OTN);
			lamdas[j]=calLamda(1,Intra[j], Intra[j+1],OTP,OTN);
//			lamdas[j]=calLamda(v1,v2,OTP,OTN);
			TP[j] += OTP;
			TN[j] += OTN;
			PN[j] += Intra[j].size() + Intra[j+1].size();
			double tall = Intra[j].size()+Intra[j+1].size();
			precs[j] = (OTP+OTN)/tall;
			cout<<i<<'\t'<<j<<'\t'<<Intra[j].size()<<'\t'<<(OTP+OTN)/tall<<'\t' << OTP<<'\t'<<OTN<<'\t'<<tall<<endl;
		}
//		SidLamda[Taxon[i][L7[0]]] = lamdas;
		SidLamda[i] = lamdas;
		SidPrec[i] = precs;
		SidNormal[i] = normalPara;
		SidNormalPrior[i] = normalPrior;
		for(int j=0;j<NL;++j)
			cerr << lamdas[j] << '\t';
		cerr << endl;
	}
	for(int i=0;i<NL;++i){
		cerr << i << '\t' << (TP[i]+TN[i])/(double)PN[i] << '\t' << TP[i] << '\t' << TN[i] << '\t' << PN[i] << endl;
	}
}
void getNRLamda(NCBI_nodes_dmp &Dmp, const string nr_mtx_info, const string nr_mtx_path, map<int,vector<double> >&SidLamda, map<int,vector<double> > &SidPrec, map<int,vector<pair<double,double> > > &SidNormal,map<int,vector<int> >&SidNormalPrior,const int LenThreshold = 0){
	const int MAX=1000000;
	char* Buf = new char[MAX];

 	const int L7[] = {4,8,12,17,21,24,28};
	vector<vector<int> >Taxon;
	vector<int>Len;
	ifstream ifs(nr_mtx_info.c_str());
	assert(!ifs.fail());
	while(!ifs.eof()){
		ifs.getline(Buf,MAX);
		int tid = -1,len=0,kmer=0;
		sscanf(Buf,"%d\t%d\t%d",&tid,&len,&kmer);
		if(tid<=0)break;
		vector<int> taxon = Dmp.getTaxo(tid);
		vector<int> tax7(7);
		for(int j=0;j<7;++j)
			tax7[j] = taxon[L7[j]];
		Taxon.push_back(tax7);
		Len.push_back(kmer);
	}
	ifs.close();
	delete[]Buf;
	int N = Taxon.size();
	cerr << "# seq: " << N << endl;

	vector<vector<double> >Mtx(N);
	for(int i=0;i<N;++i)
		Mtx[i].resize(N,0);
	ifs.open(nr_mtx_path.c_str());
	assert(!ifs.fail());
	for(int i=0;i<N;++i)
		for(int j=0;j<N;++j)
			ifs >> Mtx[i][j];
	ifs.close();
	cerr << "int Mtx: \t" << Mtx[3][0] << '\t' << Mtx[4][N-1] << endl;
	cerr << "int Mtx: \t" << Mtx[N-1][0] << '\t' << Mtx[N-1][N-1] << endl;
	for(int j=0;j<N;++j)
		Mtx[j][j] = max(Mtx[j][j],1e-1);

	const int NL = sizeof(L7)/sizeof(int);
	cerr << "NL :" << NL << endl;
/*	vector<int>TP(NL,0);
	vector<int>TN(NL,0);
	vector<int>PN(NL,0);*/
	omp_lock_t type_lock;
	omp_init_lock(&type_lock);
#pragma omp parallel for schedule(dynamic)
	for(int i=0;i<N;++i){
		omp_set_lock(&type_lock);
		if(Len[i] < LenThreshold)
			cerr <<"skipping "<<i<<"\t"<<Len[i]<<"\t"<<Taxon[i][0]<< endl;
		else cerr <<"processing taxon "<<i<<"\t"<<Len[i]<<"\t"<<Taxon[i][0]<< endl;
		omp_unset_lock(&type_lock);
		if(Len[i] < LenThreshold)
			continue;
		vector<double> Intra[NL+1];
		for(int j=0;j<N;++j){
			if(Len[j] < LenThreshold)
			continue;
			int k = 0;
			for(;k<NL && Taxon[i][k]!=Taxon[j][k];++k);
			Intra[k].push_back(Mtx[i][j]/Mtx[j][j]);
		}
		//calculate normal distribution
		vector<pair<double,double> >normalPara(NL+1);
		vector<int>normalPrior(NL+1);
		for(int j=0;j<=NL;++j){
			if(Intra[j].size()==0){
				normalPara[j] = make_pair(0,0);
				continue;
			}
			double sumX = 0, sumXX = 0;
			for(vector<double>::const_iterator itr=Intra[j].begin();itr!=Intra[j].end();++itr){
				sumX += *itr;
				sumXX += (*itr)*(*itr);
			}
			sumX /= Intra[j].size();
			normalPara[j] = make_pair(sumX,sqrt(sumXX/Intra[j].size()-sumX*sumX));
			normalPrior[j] = Intra[j].size();
		}
		/////////////////////////////////////////////////////////
		vector<double> lamdas(NL);
		vector<double> precs(NL);
		for(int j=0;j<NL;++j){
			int OTP=0,OTN=0;
			lamdas[j]=calLamda(1,Intra[j], Intra[j+1],OTP,OTN);
			double tall = Intra[j].size()+Intra[j+1].size()+1e-3;
			precs[j] = (OTP+OTN)/tall;
		}
		int tid = Taxon[i][0];
		omp_set_lock(&type_lock);
		SidLamda[tid] = lamdas;
		SidPrec[tid] = precs;
		SidNormal[tid] = normalPara;
		SidNormalPrior[tid] = normalPrior;
		omp_unset_lock(&type_lock);
	}
}
void readNRLamda(const string nr_lamda_path, map<int,vector<double> >&SidLamda, map<int,vector<double> > &SidPrec, map<int,vector<pair<double,double> > > &SidNormal,map<int,vector<int> >&SidNormalPrior,const int LenThreshold = 0){
	ifstream ifs(nr_lamda_path.c_str());
	assert(!ifs.fail());
	const int MAX = 10000;
	char Buf[MAX];
	while(!ifs.eof()){
		ifs.getline(Buf,MAX);
		if(strlen(Buf)<=1)break;
		stringstream ss(Buf);
		int Sid=0,n=0;
		ss >> Sid >> n;
		double lamda;
		for(int i=0;i<n;++i){
			ss >> lamda;
			SidLamda[Sid].push_back(lamda);
		}
		ss >> n; double prec;
		for(int i=0;i<n;++i){
			ss >> prec;
			SidPrec[Sid].push_back(prec);
		}
		ss >> n; double normal1,normal2;
		for(int i=0;i<n;++i){
			ss >> normal1 >> normal2;
			SidNormal[Sid].push_back(make_pair(normal1,normal2));
		}
		ss >> n; int prior;
		for(int i=0;i<n;++i){
			ss >> prior;
			SidNormalPrior[Sid].push_back(prior);
		}
	}
	ifs.close();
}
#endif
