#include  "NRBWTGenerator.h"
#include <dirent.h>
#include <fstream>
#include <iostream>
#include <algorithm>

#define F(x) ((x)/3+((x)%3==1?0:tb))
#define G(x) ((x)<tb?(x)*3+1:((x)-tb)*3+2)

using namespace std;
bool NRBWTGenerator::c0(unsigned *r,long long a,long long b)
{
	return r[a]==r[b]&&r[a+1]==r[b+1]&&r[a+2]==r[b+2];
}
bool NRBWTGenerator::c12(long long k,unsigned *r,long long a,long long b)
{
	if(k==2) return r[a]<r[b]||r[a]==r[b]&&c12(1,r,a+1,b+1);
	else return r[a]<r[b]||r[a]==r[b]&&wv[a+1]<wv[b+1];
}
void NRBWTGenerator::radixsort(unsigned *r,unsigned *a,unsigned *b,long long n,long long m)
{
    long long i;
    for(i=0;i<n;i++) wv[i]=r[a[i]];
    for(i=0;i<m;i++) ws[i]=0;
    for(i=0;i<n;i++) ws[wv[i]]++;
    for(i=1;i<m;i++) ws[i]+=ws[i-1];
    for(i=n-1;i>=0;i--) b[--ws[wv[i]]]=a[i];
    return;
}
void NRBWTGenerator::dc3(unsigned *r,unsigned *sa,long long n,long long m)
{
    long long i,j;unsigned *rn=r+n;unsigned *san=sa+n;long long ta=0,tb=(n+1)/3,tbc=0,p;
    r[n]=r[n+1]=0;
    for(i=0;i<n;i++) if(i%3!=0) wa[tbc++]=i;
    radixsort(r+2,wa,wb,tbc,m);
    radixsort(r+1,wb,wa,tbc,m);
    radixsort(r,wa,wb,tbc,m);
    for(p=1,rn[F(wb[0])]=0,i=1;i<tbc;i++)
        rn[F(wb[i])]=c0(r,wb[i-1],wb[i])?p-1:p++;
    if(p<tbc) 
		dc3(rn,san,tbc,p);
    else 
		for(i=0;i<tbc;i++) 
			san[rn[i]]=i;
    for(i=0;i<tbc;i++) 
		if(san[i]<tb) 
			wb[ta++]=san[i]*3;
    if(n%3==1) 
		wb[ta++]=n-1;
    radixsort(r,wb,wa,ta,m);
    for(i=0;i<tbc;i++) 
		wv[wb[i]=G(san[i])]=i;
    for(i=0,j=0,p=0;i<ta && j<tbc;p++)
        sa[p]=c12(wb[j]%3,r,wa[i],wb[j])?wa[i++]:wb[j++];
    for(;i<ta;p++) 
		sa[p]=wa[i++];
    for(;j<tbc;p++) 
		sa[p]=wb[j++];
    return;
}
long long NRBWTGenerator::inputToStr(char* input)
{
	long long Len = strlen(input);
	assert(Len < MAXN);//not allowed to be equal, because we will add an extra chES at end;
	for(long long i=0;i<Len;++i)
	{
		if(input[i] >= 'a' && input[i] <= 'z')
			str[i] = input[i]-('a'-'A');
		else str[i] = input[i];
	}
	str[Len]=chES;
	str[Len+1] = 0;
	return Len+1;
}
long long NRBWTGenerator::calSA(char* input)
{
	long long Len = inputToStr(input);
	dc3(str,SA,Len,MAXM);
	return Len;
}
long long NRBWTGenerator::calBWT(char* input)
{
	long long Len = calSA(input);
	for(long long i=0;i<Len;++i)
		if(SA[i])
			bwt[i] = input[SA[i]-1];
		else
			bwt[i] = '$';
	bwt[Len] = 0;
	return Len;
}

int NRBWTGenerator::binary_search_taxid(int sa_idx)
{
	long long min_ = 0, max_ = BinarySearchTable.size()-1;
	while(min_ < max_)
	{
		long long mid_ = (min_+max_)/2;
		if(BinarySearchTable[mid_].first <= sa_idx)
			min_ = mid_+1;
		else
			max_ = mid_;
	}
	if(BinarySearchTable[min_].first >= sa_idx)
		return min_;
//		return BinarySearchTable[min_].second;
	return BinarySearchTable.size();
}

void NRBWTGenerator::outputSA_BWT_Taxid(string outfile, long long bwtstridx )
{
	ofstream ofs(outfile.c_str(), ios_base::out|ios_base::app|ios_base::binary);
	assert(!ofs.fail());
/*	for(long long i=0;i<bwtstridx;i+=2)
		ofs << combine_chars(bwt[i],bwt[i+1]);*/
	for(long long i=0;i<bwtstridx;++i)
		ofs << bwt[i];
	ofs << endl;
	for(long long i=0;i<bwtstridx;++i)
		SA[i]=binary_search_taxid(SA[i]);
	ofs.write((char*)SA,bwtstridx*4); 
	////////////////////////////////////////////
	int* Id2Taxon = new int[BinarySearchTable.size()+1];
	Id2Taxon[0] = BinarySearchTable.size();
	for(int i=0;i<BinarySearchTable.size();++i)
		Id2Taxon[i+1] = BinarySearchTable[i].second;
	ofs.write((char*)Id2Taxon,(BinarySearchTable.size()+1)*4);
	delete[]Id2Taxon;
	////////////////////////////////////////////
	ofs.close();
}
///////////////////////////////////////////////////////////////////////////////////
bool cmpProteinTid(const pair<string,pair<int,int> >& t1, const pair<string,pair<int,int> >& t2){
	if(t1.second.second == t2.second.second)
		return t1.second.first < t2.second.first;
	return t1.second.second < t2.second.second;
}
int getLCAFromNr(int* Gi_Taxid,char* Buf,const NCBI_nodes_dmp& NodesDmp){
	int len = strlen(Buf);
	set<int> S;
	int x = atoi(Buf+4);
	if(x >= 0 && Gi_Taxid[x] > 0)
		S.insert(Gi_Taxid[x]);
	for(int i=4;i<len;++i){
		if(Buf[i]==1){
			int x = atoi(Buf+i+4);
			if(x >= 0 && Gi_Taxid[x] > 0)
				S.insert(Gi_Taxid[x]);
		}
	}
	if(S.size()==0)return -1;
	else if(S.size()==1)return *S.begin();
 	const int L7[] = {4,8,12,17,21,24,28};
	const int NL = 7;
	vector<vector<int> >V;
	for(set<int>::const_iterator itr=S.begin();itr!=S.end();++itr)
		V.push_back(NodesDmp.getTaxo(*itr));
	for(int k=0;k<NL;++k){
		bool allSame = true;
		for(int j=1;j<V.size() && allSame;++j){
			if(V[j][L7[k]] != V[0][L7[k]])
				allSame = false;
		}
		if(allSame)
			return V[0][L7[k]];
	}
	return -2;
}
void NRBWTGenerator::calBWTfromNr(const std::string NrPath,const std::string gi_taxid_nucl_path, const std::string NodesDmp_Path,std::string outfile){
	////get gi-taxid array
	const long long MAX_GI = 1<<30;
	ifstream gifs(gi_taxid_nucl_path.c_str());
	if(gifs.fail()){
		cerr << "gi_taxid_nucl file open failed: "<<gi_taxid_nucl_path << endl;
		exit(-1);
	}
	const int BufMax=30000000;
	char* Buf = new char[BufMax];
	int* Gi_Taxid = new int[MAX_GI];
	for(long long i = 0;i<MAX_GI;++i)
		Gi_Taxid[i] = -1;
	cerr << " start to load gi_taxid_nucl.dmp." << endl;
	{
		long long tgi,ttaxid,txtidx = 0;
		while(!gifs.eof())
		{
			tgi=-1,ttaxid = -1;
			gifs.getline(Buf,BufMax);
			sscanf(Buf,"%lld\t%lld",&tgi,&ttaxid);
			if(tgi>=MAX_GI)cerr<<"Array too small. current gid: "<<tgi<<"\t"<<txtidx<<"\t"<<Buf<<endl;
			else if(tgi>=0)
				Gi_Taxid[tgi] = ttaxid;
			++txtidx;
		}
		gifs.close();
		cerr << " gi_taxid_nucl.dmp is loaded. # of lines:"<<txtidx<<"; max gi:"<<tgi<<"; max taxid:"<<ttaxid << endl;
	}
	//////////////////////////////////////////////////////////////
	////get Nodes.Dmp
	NCBI_nodes_dmp NodesDmp;
	NodesDmp.init(NodesDmp_Path);
	//////////////////////////////////////////////////////////////
	////get string
	vector<pair<string,pair<int,int> > >proteindata;
	char* Protein = new char[MAXN];
	unsigned ProteinIdx = 0;
	int current_taxid = 0;
	int current_gi1 = 0;
	ifstream ifs(NrPath.c_str());
	if(ifs.fail()){
		cerr << "file open failed! " << NrPath << endl;
		exit(-1);
	}
 	const int L7[] = {4,8,12,17,21,24,28};
	const int NL = 7;
	assert(!ifs.fail());
	while(!ifs.eof()){
		Buf[0] = 0;
		ifs.getline(Buf,BufMax);
		if(Buf[0] == '>'){
			if(current_taxid > 0)
				proteindata.push_back(make_pair(string(Protein),make_pair(current_gi1,current_taxid)));
			current_gi1 = atoi(Buf+4);
			current_taxid = getLCAFromNr(Gi_Taxid, Buf, NodesDmp);
			vector<int> taxon = NodesDmp.getTaxo(current_taxid);
			if(taxon[L7[6]]<1 || taxon[L7[6]]==2759)
				current_taxid=0;
			ProteinIdx = 0;
		}else{
			strcpy(Protein+ProteinIdx,Buf);
			ProteinIdx += strlen(Buf);
		}
	}
	ifs.close();
	if(current_taxid > 0)
		proteindata.push_back(make_pair(string(Protein),make_pair(current_gi1,current_taxid)));
	sort(proteindata.begin(),proteindata.end(),cmpProteinTid);
	////////////////////////////////////////////////////////////
	////get bwt
	outfile += ".idx";
	system((string("rm ")+outfile).c_str());
	BinarySearchTable.clear();
	long long bwtstridx = 0;
	cerr << "# taxa: " << proteindata.size() << endl;
	////////////////////////////////////////////////////////////
	const char InsertBase = '9';
	vector<string> usedFileInfo;
	for(int i=0;i<proteindata.size();){
		int j = i+1; 
		ProteinIdx = 0;
		////////////////////////////////////////////////////
	/*	for(long long tlen = proteindata[i].first.length()+1;j < proteindata.size() && proteindata[j].second == proteindata[i].second && tlen < MAXN; tlen += (proteindata[j].first.length()+1),++j);
		for(int k=i;k<j;++k){
			strcpy(Protein+ProteinIdx,proteindata[k].first.c_str());
			ProteinIdx += proteindata[k].first.length();
			Protein[ProteinIdx++] = InsertBase;
		}*/
		strcpy(Protein+ProteinIdx,proteindata[i].first.c_str());
		ProteinIdx += proteindata[i].first.length();
		Protein[ProteinIdx++] = InsertBase;
		////////////////////////////////////////////////////
		if(bwtstridx + ProteinIdx >= MAXN-1){
			long long bwtlen=calBWT(inputstr);
			outputSA_BWT_Taxid(outfile, bwtlen);
			cerr << "t2 ouput taxa: " << i << '\t' << bwtlen << endl;
			bwtstridx = 0;
			BinarySearchTable.clear();
		}
		Protein[ProteinIdx] = 0;
		strcpy(inputstr+bwtstridx,Protein);
		bwtstridx += ProteinIdx;
		BinarySearchTable.push_back(make_pair(bwtstridx,proteindata[i].second.second));

		char ginfo[100];
		sprintf(ginfo,"%d\t%d\t%d\t%d\0",proteindata[i].second.second,ProteinIdx,i,proteindata[i].second.first);
		usedFileInfo.push_back(ginfo);
		i = j;
	}
	if(bwtstridx > 0){
		long long bwtlen = calBWT(inputstr);
		outputSA_BWT_Taxid(outfile, bwtlen);
		cerr << "t4 ouput taxa: " << bwtlen << endl;
	}
	/////////////////////////////////////////////////
	ofstream ofs((outfile+".info").c_str());
	if(!ofs.fail()){
		for(vector<string>::const_iterator itr= usedFileInfo.begin();itr!=usedFileInfo.end();++itr)
			ofs << *itr << endl;
	}
	ofs.close();
	delete[] Buf;
	delete[] Protein;
	delete[] Gi_Taxid;
}
