#include <iostream>
#include <fstream>
#include <cstdio>
#include <string>
#include <cstring>
#include <cstdlib>
#include <cassert>
#include <omp.h>
#include "BWTs.h"
using namespace std;
void usage(){
	cerr << "calMTXofCG bwtofcg.idx (20-60)" << endl;
	exit(-1);
}
struct SeqInfoUnit{
	int gid,tid,len;
	string path;
	SeqInfoUnit(int g,int t,int l, string s):gid(g),tid(t),len(l),path(s){}
};
vector<SeqInfoUnit> SeqInfo;

const int MAX = 1000;
const int MAXL = 50000000;
int main(int argc, char* argv[])
{
	if(argc < 2)
		usage();

	ifstream ifs((string(argv[1])+".info").c_str());
	if(ifs.fail()){
		cerr << "Open failed: " << argv[1] << endl;
		exit(-1);
	}
	while(!ifs.eof()){
		int gid,tid,len=0;
		string path;
		ifs >> gid >> tid >> len >> path;
		if(len<=0)break;
		assert(len < MAXL);
		SeqInfo.push_back(SeqInfoUnit(gid,tid,len,path));
	}
	ifs.close();
	int N = SeqInfo.size();
	cerr << "#. seqs: " << N << endl;
	cerr << SeqInfo[N-1].path << endl;

	vector<int>substrLength;
	for(int THIS_PARA_KMER = 20; THIS_PARA_KMER <= 60; THIS_PARA_KMER+=5)
		substrLength.push_back(THIS_PARA_KMER);
	int Nsubstr = substrLength.size();
	int SHORTEST_PARA_KMER = substrLength[0];
	int LONGEST_PARA_KMER = substrLength[Nsubstr-1];
	int*** MTX = new int**[Nsubstr];
	for(int i=0;i<Nsubstr;++i){
		MTX[i]=new int*[N];
		for(int j=0;j<N;++j){
			MTX[i][j] = new int[N];
			for(int k=0;k<N;++k)
				MTX[i][j][k] = 0;
		}
	}

	omp_lock_t type_lock;
	omp_init_lock(&type_lock);

	BWTs bwts(argv[1]);
	cerr << "bwts loaded" << endl;
	ofstream ofs((string(argv[1])+".mtx").c_str());
#pragma omp parallel for schedule(dynamic)
		for(int i=0;i<N;++i){
			ifstream ifsi(SeqInfo[i].path.c_str());
			if(ifsi.fail()){
				omp_set_lock(&type_lock);
				cerr << "Open failed: " << i << '\t' << SeqInfo[i].path << endl;
				omp_unset_lock(&type_lock);
				continue;
			}
			char Buf[MAX];
			string Seq;
			ifsi.getline(Buf,MAX);
			while(!ifsi.eof()){
				ifsi.getline(Buf,MAX);
				Seq += Buf;
	//			strcpy(Seq+charidx,Buf);
	//			charidx += strlen(Buf);
			}
			ifsi.close();
			omp_set_lock(&type_lock);
			cerr <<"processed: " << i << "\t" << SeqInfo[i].len << '\t' << Seq.length() << endl;
			omp_unset_lock(&type_lock);
	
			string str = Seq;
			set<pair<string,string> >kmers;
			{
				string rstr = str;
				for(int i=0;i<str.length();++i)
				{
					char ch = str[str.length()-1-i];
					switch(ch)
					{
						case 'A':rstr[i]='T';break;
						case 'C':rstr[i]='G';break;
						case 'G':rstr[i]='C';break;
						case 'T':rstr[i]='A';break;
						default:rstr[i]='N';
					}
				}
				int strend = (str.length()-(LONGEST_PARA_KMER-1));
				for(int i=0;i<strend;++i){
					string k1 = str.substr(i,LONGEST_PARA_KMER);
					string k2 = rstr.substr(str.length()-LONGEST_PARA_KMER-i,LONGEST_PARA_KMER);
					if(k1>k2)swap(k1,k2);
					kmers.insert(make_pair(k1,k2));
				}
				int strend2 = (str.length()-(SHORTEST_PARA_KMER-1));
				for(int i=strend;i<strend2;++i){
					string k1 = str.substr(i);
					string k2 = rstr.substr(0,str.length()-i);
					if(k1>k2)swap(k1,k2);
					kmers.insert(make_pair(k1,k2));
				}
			}
			for(set<pair<string,string> >::const_iterator itr = kmers.begin();itr!=kmers.end();++itr){
				vector<set<pair<int,int> > >gross_cand_taxid1;
				bwts.substrSearch(itr->first, substrLength, gross_cand_taxid1);
				vector<set<pair<int,int> > >gross_cand_taxid2;
				bwts.substrSearch(itr->second, substrLength, gross_cand_taxid2);
				for(int j=0;j<Nsubstr;++j){
					set<int>sid_cand;
					for(set<pair<int,int> >::const_iterator itr2=gross_cand_taxid1[j].begin();itr2!=gross_cand_taxid1[j].end();++itr2)
						sid_cand.insert(itr2->first);
					for(set<pair<int,int> >::const_iterator itr2=gross_cand_taxid2[j].begin();itr2!=gross_cand_taxid2[j].end();++itr2)
						sid_cand.insert(itr2->first);
					for(set<int>::const_iterator itr2=sid_cand.begin();itr2!=sid_cand.end();++itr2)
						++MTX[j][i][*itr2];
				}
			}
			if(i==10){
				for(int l=0;l<Nsubstr;++l){
					for(int i=0;i<N;++i){
						for(int j=0;j<N;++j)
							cout << MTX[l][i][j] << '\t';
								cout << endl;
					}
				}
			}
		}
	for(int l=0;l<Nsubstr;++l){
		for(int i=0;i<N;++i){
			for(int j=0;j<N;++j)
				ofs << MTX[l][i][j] << '\t';
			ofs << endl;
		}
	}
	ofs.close();
	return 0;
}
